//************************************************************************
//This module contains fitness case and functions to compute raw fitness *
//as well as standardized fitness for Symbolic Regression Problem        *
//Programmed by NXH.                                                     *
//************************************************************************
#ifndef __SYMBOLICREG__
#define __SYMBOLICREG__

#include<iostream.h>
#include<string.h>
#include<stdio.h>
#include<assert.h>
#include<math.h>
#include<stdlib.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
#include "adtable.h"
#include "dtree.h"

#define NUMFITCASE 20
#define INFPLUS  exp(700)
#define INFMINUS -exp(700)
//#define EXPOVL 650
#define PVAL(y) ((y==HUGE_VAL)?(INFPLUS):((y==-HUGE_VAL)?(-INFMINUS):(y)))
//****************************************
//Fitness cases for symbolic regression  *
//They are sampled between 0  and 1      *
//Seed for sampling 12345                *
//****************************************
struct _sample{
		 double x;
		 double y;
		};
typedef struct _sample sample;
sample fitcase[NUMFITCASE];
BOOL SuccPredicate=FALSE;
void SetFitCase()
{
 //FILE *fp;
  double x;
  int i;
//  char filename[50];
//  cout<<endl<<"Input Data File Name:";
//  cin>>filename;
//  fp=fopen(filename,"rt");assert(fp!=NULL);
  for(i=0;i<NUMFITCASE;i++)
   {
    // fscanf(fp,"%lf%lf",&x,&y);
     fitcase[i].x=2*Next_Double()-1;
     x=fitcase[i].x;
     fitcase[i].y=x*x*x*x+x*x*x+x*x+x;
   }
 // fclose(fp);
}
//********************************************************************
//This function compute the function (in tree form) with given input *
//********************************************************************
double Compute(node *st,double x)
{
 node *p;
 double l,r;
 if (!strcmp(st->name,"X"))
     return x;
 else //st->name="EXP"
  {
    p=st->children;
    if (!strcmp(p->name,"EXP"))
    {
     l=Compute(p,x);//compute the left branch
     p=p->sibling;
     r=Compute(p->sibling,x);//compute the right branch
     p=p->children;//get the operator
     if (!strcmp(p->name,"add"))
       return PVAL(r+l);
     else
      if (!strcmp(p->name,"sub"))
	return PVAL(l-r);
	else
	 if (!strcmp(p->name,"mul"))
	  return PVAL(r*l);
	  else
	    if (r!=0)   //protected division
	     return PVAL(l/r);
	     else
	      return 1;
    }
   else
    if (!strcmp(p->name,"PRE"))
     {
       r=Compute(p->sibling,x); //compute exp inside function
       p=p->children;
       if (!strcmp(p->name,"sin"))
	return sin(r);
       else
	if(!strcmp(p->name,"cos"))
	 return cos(r);
	  else
	  if(!strcmp(p->name,"ep"))
	      return PVAL(exp(r));
	   else
	     if (r!=0)   //protected logarithm
	      return log(fabs(r));
	       else
		return 0;
     }
    else //VAR
      return Compute(p->children,x);
  }
}
//*********************************************
//Compute the raw fitness for an invidual     *
//*********************************************
double ComputeRF(stree st)
{
  double sum=0,t;
  int i;
  int hit;
  hit=0;
  for(i=0;i<NUMFITCASE;i++)
   {
     t=PVAL(fabs(Compute(st.root,fitcase[i].x)-fitcase[i].y));
     if(t<=0.01)
       hit++;
     sum+=t;
   }
   if(hit==NUMFITCASE)
     SuccPredicate=TRUE;
   return sum;
}
void BuildVocab()
{
 vocab=NULL;
 lexnum=0;
}
//****************************************************************
//This function initializes elementary trees and adjoining table *
//****************************************************************
void LoadEtree()
{
 cout<<endl<<"Loading elementary trees...";
 etree=new stree[14];assert(etree!=NULL);
 etree[0]=InitTree("EXP(VAR(X))");
 etree[1]=InitTree("EXP(EXP(VAR(X)),OP(add),EXP~)");
 etree[2]=InitTree("EXP(EXP(VAR(X)),OP(mul),EXP~)");
 etree[3]=InitTree("EXP(EXP(VAR(X)),OP(sub),EXP~)");
 etree[4]=InitTree("EXP(EXP(VAR(X)),OP(div),EXP~)");
 etree[5]=InitTree("EXP(EXP~,OP(add),EXP(VAR(X)))");
 etree[6]=InitTree("EXP(EXP~,OP(mul),EXP(VAR(X)))");
 etree[7]=InitTree("EXP(EXP~,OP(sub),EXP(VAR(X)))");
 etree[8]=InitTree("EXP(EXP~,OP(div),EXP(VAR(X)))");
 etree[9]=InitTree("EXP(PRE(sin),EXP~)");
 etree[10]=InitTree("EXP(PRE(cos),EXP~)");
 etree[11]=InitTree("EXP(PRE(ep),EXP~)");
 etree[12]=InitTree("EXP(PRE(lg),EXP~)");
 numetree=13;
 numalpha=1;
 numbeta=12;
 cout<<"Done!";
 cout<<endl<<"Building Vocabulary...";
 BuildVocab();
 cout<<"done";
 cout<<endl<<"Buiding adjoining table...";
 entrynum=0;
 BuildAdTable();
 cout<<"Done!";
}
node *Stree2ParseTree(node *st)
{
 node *p,*q,*t,*t1,*t2;
 if(!strcmp(st->name,"EXP"))
  {
   p=st->children;
   if(!strcmp(p->name,"EXP"))//binary oprators
    {
     //get the oprators
     q=p->sibling->children;
     t=MakeNode(q->name);
     t1=Stree2ParseTree(p);
     t2=Stree2ParseTree(p->sibling->sibling);
     t->children=t1;
     t1->sibling=t2;
     t2->sibling=NULL;
     return t;
    }
    else
     if(!strcmp(p->name,"PRE"))
     {
      q=p->children;
      t=MakeNode(q->name);
      t1=Stree2ParseTree(p->sibling);
      t->children=t1;
      t1->sibling=NULL;
      return t;
     }
    else //VAR
     return Stree2ParseTree(p->children);
  }
 else //"X"
  {
    t=MakeNode(st->name);
    return t;
  }

}
#endif
